#ifndef MSP_DATAFILE_COLLECTION_H_
#define MSP_DATAFILE_COLLECTION_H_

#include <msp/core/maputils.h>
#include <msp/core/meta.h>
#include <msp/core/refptr.h>
#include "collectionsource.h"
#include "loader.h"

/* XXX This file is a big mess with too many things in it.  However, the
dependencies between those things make it difficult to split up. */

namespace Msp {
namespace DataFile {

/**
Helper struct to determine whether a Loader has a Collection typedef.
*/
template<typename T>
struct NeedsCollection
{
	struct Yes { char c[2]; };
	struct No { char c; };

	template<typename U>
	static Yes f(typename U::Collection *);
	template<typename U>
	static No f(...);

	enum { value = (sizeof(f<T>(0))==sizeof(Yes)) };
};

class CollectionItemTypeBase;

template<typename T>
class CollectionItemType;

/**
A collection of objects that can be loaded from a datafile.  Each object is
identified by a name, which must be unique across the entire collection.

While this class can be instantiated by itself and used for storing objects,
loading requires that a subclass defines the supported types.  See the add_type
method for details.

Collections can have sources for loading objects on demand.  Automatic loading
only works on a non-const Collection.  See class CollectionSource for details.
*/
class Collection
{
public:
	/**
	Loads objects into a Collection.  Automatically picks up keywords from
	defined item types.
	*/
	class Loader: public DataFile::Loader
	{
		template<typename T> friend class CollectionItemType;

	private:
		Collection &coll;

	public:
		Loader(Collection &);
		Collection &get_object() const { return coll; }
	private:
		template<typename T, typename S>
		void item(const std::string &n)
		{
			RefPtr<T> it = new T;
			ItemLoader<T> ldr(*it, coll);
			load_sub_with(ldr);
			coll.add<S>(n, it.get());
			it.release();
		}
	};

protected:
	template<typename T, bool = NeedsCollection<typename T::Loader>::value>
	class ItemLoader;

private:
	typedef std::map<std::string, Variant> ItemMap;
	typedef std::list<CollectionItemTypeBase *> TypeList;
	typedef std::list<CollectionSource *> SourceList;

	TypeList types;
	ItemMap items;
	SourceList sources;

	Collection(const Collection &);
	Collection &operator=(const Collection &);
public:
	Collection() { }
	virtual ~Collection();

	/** Adds an object into the collection.  The name must not pre-exist.  The
	collection takes ownership of the object. */
	template<typename T>
	void add(const std::string &name, T *item)
	{
		if(!item)
			throw std::invalid_argument("Collection::add(item)");

		insert_unique(items, name, RefPtr<typename RemoveConst<T>::Type>(item));
	}

	/// Gets a typed object from the collection.
	template<typename T>
	T &get(const std::string &name) const
	{
		return *get_item(items, name).value<RefPtr<typename RemoveConst<T>::Type> >();
	}

	/** Gets a typed object from the collection.  If the name is not found,
	automatic creation with the type's creator function (if defined) or from
	sources (if present) is attempted. */
	template<typename T>
	T &get(const std::string &);

private:
	template<typename T>
	void collect_items(std::list<T *> *objects, std::list<std::string> *names, std::list<std::string> *future_names) const
	{
		typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;

		for(ItemMap::const_iterator i=items.begin(); i!=items.end(); ++i)
			if(i->second.check_type<RPNCT>())
			{
				if(objects)
					objects->push_back(i->second.value<RPNCT>().get());
				if(names)
					names->push_back(i->first);
			}

		if(future_names)
			if(CollectionItemTypeBase *type = get_type<T>())
			{
				for(SourceList::const_iterator i=sources.begin(); i!=sources.end(); ++i)
				{
					std::list<std::string> available_names = (*i)->get_names(*type);
					for(std::list<std::string>::iterator j=available_names.begin(); j!=available_names.end(); ++j)
						if(!items.count(*j))
							future_names->push_back(*j);
				}
			}
	}

public:
	/// Returns a list of the names of objects of one type in the collection.
	template<typename T>
	std::list<std::string> get_names() const
	{
		std::list<std::string> result;
		collect_items<T>(0, &result, 0);
		return result;
	}

	/** Returns a list of the names of objects of one type in the collection or
	available from sources. */
	template<typename T>
	std::list<std::string> get_names()
	{
		std::list<std::string> result;
		collect_items<T>(0, &result, &result);
		return result;
	}

	/// Returns a list of objects of one type in the collection.
	template<typename T>
	std::list<T *> get_list() const
	{
		std::list<T *> result;
		collect_items<T>(&result, 0, 0);
		return result;
	}

	/** Returns a list of objects of one type, loading them from sources if
	necessary. */
	template<typename T>
	std::list<T *> get_list()
	{
		std::list<T *> result;
		std::list<std::string> future;
		collect_items<T>(&result, 0, &future);
		for(std::list<std::string>::iterator i=future.begin(); i!=future.end(); ++i)
			result.push_back(&get<T>(*i));
		return result;
	}

private:
	template<typename T>
	unsigned get_status(const std::string &name) const
	{
		ItemMap::const_iterator i = items.find(name);
		if(i==items.end())
		{
			if(CollectionItemTypeBase *type = get_type<T>())
			{
				for(SourceList::const_iterator j=sources.begin(); j!=sources.end(); ++j)
					if((*j)->is_loadable(*type, name))
						return 2;
			}
			return 0;
		}

		typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;
		if(!i->second.check_type<RPNCT>())
			return 0;

		return 1;
	}

public:
	/// Checks whether a typed object exists in the collection.
	template<typename T>
	bool contains(const std::string &name) const
	{ return get_status<T>(name)==1; }

	/** Checks whether a typed object exists in the collection or is loadable
	from a source. */
	template<typename T>
	bool contains(const std::string &name)
	{ return get_status<T>(name)>0; }

	/// Returns the name of an item in the collection.
	template<typename T>
	const std::string &get_name(T *d) const
	{
		typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;

		for(ItemMap::const_iterator i=items.begin(); i!=items.end(); ++i)
			if(i->second.check_type<RPNCT>())
				if(i->second.value<RPNCT>().get()==d)
					return i->first;
	
		// XXX Need better exception class
		throw std::runtime_error("Item not found in collection");
	}

protected:
	/** Adds a type to the collection.  The returned descriptor object reference
	can be used to define how objects of that type can be loaded. */
	template<typename T>
	CollectionItemType<T> &add_type();

	/** Returns the descriptor for a type, or null if one isn't defined. */
	template<typename T>
	CollectionItemTypeBase *get_type() const;

	void add_source(CollectionSource &);
};

template<typename T>
class Collection::ItemLoader<T, false>: public T::Loader
{
public:
	ItemLoader(T &o, Collection &):
		T::Loader(o)
	{ }
};

template<typename T>
class Collection::ItemLoader<T, true>: public T::Loader
{
public:
	ItemLoader(T &o, Collection &c):
		T::Loader(o, dynamic_cast<typename T::Loader::Collection &>(c))
	{ }
};


class CollectionItemTypeBase
{
protected:
	class TagBase
	{
	protected:
		TagBase() { }
	public:
		virtual ~TagBase() { }
	};

	template<typename T>
	class Tag: public TagBase
	{ };

	std::string kwd;
	std::vector<std::string> suffixes;
	TagBase *tag;

	CollectionItemTypeBase();
public:
	virtual ~CollectionItemTypeBase();

	void set_keyword(const std::string &);
	const std::string &get_keyword() const { return kwd; }
	void add_suffix(const std::string &);
	bool match_name(const std::string &) const;
	virtual void add_to_loader(Collection::Loader &) const = 0;
	virtual bool can_create() const = 0;
	virtual void create_item(Collection &, const std::string &) const = 0;
	virtual void load_item(Collection &, Parser &, const std::string &) const = 0;

	template<typename T>
	bool check_type() const
	{ return dynamic_cast<Tag<T> *>(tag); }
};


/**
Describes a type of item that can be loaded by a Collection.  These are created
by Collection::add_type.
*/
template<typename T>
class CollectionItemType: public CollectionItemTypeBase
{
private:
	class CreatorBase
	{
	protected:
		CreatorBase() { }
	public:
		virtual ~CreatorBase() { }

		virtual T *create(Collection &, const std::string &) const = 0;
	};

	template<typename C>
	class Creator: public CreatorBase
	{
	public:
		typedef T *(C::*FuncPtr)(const std::string &);

	private:
		FuncPtr func;

	public:
		Creator(FuncPtr f): func(f) { }

		virtual T *create(Collection &coll, const std::string &name) const
		{ return (static_cast<C &>(coll).*func)(name); }
	};

	class StoreBase
	{
	protected:
		StoreBase() { }
	public:
		virtual ~StoreBase() { }

		virtual void store(Collection &, const std::string &, T *) = 0;

		virtual void add_to_loader(Collection::Loader &, const std::string &) = 0;
	};

	template<typename S>
	class Store: public StoreBase
	{
	public:
		virtual void store(Collection &coll, const std::string &name, T *obj)
		{ coll.add(name, static_cast<S *>(obj)); }

		virtual void add_to_loader(Collection::Loader &loader, const std::string &kwd)
		{ loader.add(kwd, &Collection::Loader::item<T, S>); }
	};

	CreatorBase *creat;
	StoreBase *store;

public:
	CollectionItemType():
		creat(0), store(new Store<T>)
	{ tag = new Tag<T>; }

	~CollectionItemType()
	{
		delete creat;
		delete store;
	}

	/** Sets a datafile keyword for this item type.  The Collection's loader
	will accept a statement with this keyword and a single string argument - the
	item's name. */
	CollectionItemType &keyword(const std::string &k)
	{
		set_keyword(k);
		return *this;
	}

	/** Adds a suffix that is used to match names when looking for future
	objects.  There is no implied separator; a name matches if it ends with the
	suffix.  If a keyword is defined before any suffixes, then "."+keyword is
	added as a suffix. */
	CollectionItemType &suffix(const std::string &s)
	{
		add_suffix(s);
		return *this;
	}

	/** Attaches a creator function to this item type.  If an item is not found
	in the Collection, the creator function for its type is called to create it.
	The function must be a member of the Collection subclass containing the
	type.  It must return the created object, or null if it could not be
	created.  It's also permissible to load the item via other means and then
	return null. */
	template<typename C>
	CollectionItemType &creator(T *(C::*func)(const std::string &))
	{
		delete creat;
		creat = new Creator<C>(func);
		return *this;
	}

	/** Specifies the storage type for items of this type.  It must be a base
	class of the actual type.  */
	template<typename S>
	CollectionItemType &store_as()
	{
		delete tag;
		tag = new Tag<S>;
		delete store;
		store = new Store<S>;
		return *this;
	}

	virtual void add_to_loader(Collection::Loader &loader) const
	{ store->add_to_loader(loader, kwd); }

	virtual bool can_create() const
	{ return creat!=0; }

	virtual void create_item(Collection &coll, const std::string &name) const
	{
		if(!creat)
			throw std::runtime_error("no creator");
		T *obj = creat->create(coll, name);
		if(obj)
			store->store(coll, name, obj);
	}

	virtual void load_item(Collection &coll, Parser &parser, const std::string &name) const
	{
		RefPtr<T> obj = new T;
		Collection::ItemLoader<T> ldr(*obj, coll);
		ldr.load(parser);
		store->store(coll, name, obj.get());
		obj.release();
	}
};


template<typename T>
T &Collection::get(const std::string &name)
{
	typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;

	ItemMap::iterator i = items.find(name);
	if(i!=items.end())
		return *i->second.value<RPNCT>();

	if(CollectionItemTypeBase *type = get_type<T>())
	{
		bool loaded = false;
		if(type->can_create())
		{
			type->create_item(*this, name);
			loaded = items.count(name);
		}
		for(SourceList::iterator j=sources.begin(); (!loaded && j!=sources.end()); ++j)
		{
			(*j)->load(*this, *type, name);
			loaded = items.count(name);
		}
	}

	return *get_item(items, name).value<RPNCT>();
}

template<typename T>
CollectionItemType<T> &Collection::add_type()
{
	CollectionItemType<T> *type = new CollectionItemType<T>;
	types.push_back(type);
	return *type;
}

template<typename T>
CollectionItemTypeBase *Collection::get_type() const
{
	for(TypeList::const_iterator j=types.begin(); j!=types.end(); ++j)
		if((*j)->check_type<typename RemoveConst<T>::Type>())
			return *j;
	return 0;
}

} // namespace DataFile
} // namespace Msp

#endif
