#include <msp/strings/format.h>
#include "dataerror.h"
#include "loader.h"
#include "type.h"

using namespace std;

namespace {

template<typename T>
struct Set
{
	T &ref;
	T orig;

	Set(T &r, const T &v): ref(r), orig(r) { r = v; }
	~Set() { ref = orig; }
};

bool signature_match(char s, char a)
{
	if(s==a)
		return true;
	if(s==Msp::DataFile::IntType::signature && a==Msp::DataFile::FloatType::signature)
		return true;
	return false;
}

bool signature_match(const string &st_sig, const string &act_sig)
{
	if(act_sig=="*")
		return true;
	else if(act_sig.size()==2 && act_sig[1]=='*')
	{
		for(string::const_iterator i=st_sig.begin(); i!=st_sig.end(); ++i)
			if(*i!=act_sig[0])
				return false;

		return true;
	}
	else if(st_sig.size()==act_sig.size())
	{
		for(unsigned i=0; i<st_sig.size(); ++i)
			if(!signature_match(st_sig[i], act_sig[i]))
				return false;

		return true;
	}
	else
		return false;
}

}


namespace Msp {
namespace DataFile {

class unknown_keyword: public runtime_error
{
public:
	unknown_keyword(const std::string &k):
		runtime_error(k)
	{ }

	virtual ~unknown_keyword() throw() { }
};


class invalid_signature: public runtime_error
{
public:
	invalid_signature(const std::string &k, const std::string & s):
		runtime_error(format("%s %s", k, s))
	{ }

	virtual ~invalid_signature() throw() { }
};


Loader::Loader():
	cur_st(0),
	check_sub_loads(false)
{ }

Loader::~Loader()
{
	for(ActionMap::iterator i=actions.begin(); i!=actions.end(); ++i)
		delete i->second;
}

void Loader::load(Parser &p)
{
	while(p)
	{
		Statement st = p.parse();
		if(st.valid)
			load_statement(st);
	}
	finish();
}

void Loader::load(const Statement &st)
{
	for(list<Statement>::const_iterator i=st.sub.begin(); i!=st.sub.end(); ++i)
		load_statement(*i);
	finish();
}

void Loader::load_statement(const Statement &st)
{
	Set<const Statement *> set_cst(cur_st, &st);

	try
	{
		LoaderAction *act = find_action(ActionKey(st.keyword, st.get_signature()));
		if(act)
		{
			sub_loaded = false;
			act->execute(*this, st);
			if(check_sub_loads && !st.sub.empty() && !sub_loaded)
				throw logic_error("substatements ignored");
		}
	}
	catch(const data_error &)
	{
		throw;
	}
	catch(const exception &e)
	{
		throw data_error(st.source, st.line, e);
	}
}

void Loader::load_sub_with(Loader &ldr)
{
	if(!cur_st)
		throw logic_error("no current statement");

	ldr.load(*cur_st);
	sub_loaded = true;
}

void Loader::add(const string &kwd, LoaderAction *act)
{
	ActionKey key(kwd, (act ? act->get_signature() : "*"));
	ActionMap::iterator i = actions.find(key);
	if(i!=actions.end())
	{
		delete i->second;
		i->second = act;
	}
	else
		actions[key] = act;
}

LoaderAction *Loader::find_action(const ActionKey &key) const
{
	ActionMap::const_iterator begin = actions.lower_bound(ActionKey(key.keyword, string()));
	ActionMap::const_iterator end = actions.upper_bound(ActionKey(key.keyword, "~"));

	if(begin==end)
		throw unknown_keyword(key.keyword);

	for(ActionMap::const_iterator i=begin; i!=end; ++i)
		if(signature_match(key.signature, i->first.signature))
			return i->second;

	throw invalid_signature(key.keyword, key.signature);
}

const string &Loader::get_source() const
{
	if(!cur_st)
		throw logic_error("no current statement");
	return cur_st->source;
}


Loader::ActionKey::ActionKey(const string &k, const string &s):
	keyword(k),
	signature(s)
{ }

bool Loader::ActionKey::operator<(const ActionKey &other) const
{
	if(keyword!=other.keyword)
		return keyword<other.keyword;
	return signature<other.signature;
}

} // namespace DataFile
} // namespace Msp
